# Copyright (c) 2016-2021 The Regents of the University of Michigan
# Part of fresnel, released under the BSD 3-Clause License.

"""Version and build information.

Attributes:
    version (str): fresnel package version, following semantic versioning.
"""

version = "0.13.4"
